(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Infinite Intervals\<close>

theory InfiniteInterval
 imports
   Interval 
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="cvc4 z3 e spass vampire " ]
*)

text \<open>
 An infinite interval is a mapping from the natural numbers to a particular type.
 This is similar as the theory \<open>Omega_Words_Fun\<close> of the Isabelle/HOL distribution.
 The difference is that our version has no empty (no symbols) word. This is needed
 as an finite interval has at least one state. So we have to adapt the  definition of 
  \<open>conc\<close> and \<open>upt\<close>. We also define the usual \<open>isuffix\<close>, \<open>iprefix\<close> and \<open>subinterval\<close> 
 on infinite intervals. 
\<close>

subsection \<open>Definitions\<close>

type_synonym 
 'a infinterval = "nat \<Rightarrow> 'a"

type_synonym
  infiniteindex = "nat infinterval"

definition 
 conc :: "['a interval, 'a infinterval] \<Rightarrow> 'a infinterval" 
where " conc w x = (\<lambda>n. if n \<le> intlen w then nth w n else x (n  - intlen w -1))"

primrec upt :: "nat \<Rightarrow> nat \<Rightarrow> nat interval" ("(1[_..\<le>/_'])")
 where
    upt_0  : "[i..\<le>0]       = \<langle>0\<rangle>"
  | upt_Suc: "[i..\<le>(Suc j)] = (if i \<le> j then [i..\<le>j] \<ominus> \<langle>(Suc j)\<rangle> else \<langle>(Suc j)\<rangle> )"


definition
 isuffix :: "[nat, 'a infinterval] \<Rightarrow> 'a infinterval"
 where "isuffix k x = (\<lambda>n. x (k+n))"

definition 
 subinterval :: "'a infinterval \<Rightarrow> nat \<Rightarrow> nat \<Rightarrow> 'a interval"
 where
  "subinterval w i j = map w [i..\<le>j]"

definition
 iprefix :: "nat \<Rightarrow> 'a infinterval \<Rightarrow> 'a interval"
 where
  "iprefix n w \<equiv> subinterval w 0 n"

definition infinite_index_sequence :: " nat \<Rightarrow> infiniteindex \<Rightarrow> bool" where
  "infinite_index_sequence x idx \<equiv> (idx 0 = x) \<and> (\<forall> n.  idx n < idx (Suc n))"

subsection \<open>Lemmas\<close>

subsubsection \<open>upt\<close>

lemma upt_rec[code]:  
 "[i..\<le>j] = (if i<j then i \<odot> [Suc i..\<le>j] else \<langle>j\<rangle>)"
by (induct j) auto

lemma upt_conv_st [simp]:
assumes "j< i"
shows   "  [i..\<le>j] = \<langle>j\<rangle>"
using assms
by (metis InfiniteInterval.upt.simps(1) InfiniteInterval.upt.simps(2) Suc_leD less_Suc_eq_0_disj not_le)

lemma upt_same:
 "[i..\<le>i] = \<langle>i\<rangle>"
by (metis InfiniteInterval.upt.simps(1) InfiniteInterval.upt.simps(2) 
           less_Suc_eq less_Suc_eq_0_disj not_le)

lemma upt_eq_st_conv[simp]:
 "([i..\<le>j] = \<langle>j\<rangle>) = (j\<le>i)"
by (simp add: InfiniteInterval.upt_rec)

lemma upt_eq_cons_conv:
 "([i..\<le>j] = x\<odot>xs) = (i<j \<and> i=x \<and> [Suc i..\<le>j] = xs)"
using InfiniteInterval.upt_rec by (induct j arbitrary: x xs, simp, auto)

lemma upt_suc_append:
assumes "i \<le> j"
shows  " [i..\<le>(Suc j)] = [i..\<le>j]\<ominus>\<langle>(Suc j)\<rangle>"
using assms by simp

lemma upt_conv_cons:
assumes "i<j"
shows   "  [i..\<le>j] = i \<odot> [(Suc i)..\<le>j]"
using assms by (simp add: upt_rec)

lemma upt_conv_cons_cons:
 "( m \<odot> n \<odot> ns = [m..\<le>q]) = (n \<odot> ns = [(Suc m)..\<le>q])"
proof (cases "m \<le> q")
case True
then show ?thesis by (simp add: InfiniteInterval.upt_rec)
next
case False
then show ?thesis by auto
qed

lemma upt_add_eq_append:
assumes "i \<le> j \<and> k>0"
shows   "  [i..\<le>j+k] = [i..\<le>j]\<ominus>[Suc j..\<le>j+k]"
using assms 
proof
 (induct k)
 case 0
 then show ?case by blast
 next
 case (Suc k)
 then show ?case using Suc_less_eq le_simps(2) by auto
qed

lemma upt_length:
 "intlen [i..\<le>j] = j - i"
by (induct j) (auto simp add: Suc_diff_le)

lemma test:
 "k \<le> intlen [i..\<le>i + k]"
by (simp add: upt_length)

lemma upt_nth_help:
 "Interval.nth [i..\<le>i + k] k = i + k"
by  (induct k arbitrary: i, simp add: upt_same, 
     metis InfiniteInterval.upt_rec add_Suc_shift interval_nth_Suc less_add_same_cancel1 
           zero_less_Suc)

lemma upt_nth:
assumes "i + k \<le> j"
shows   " (nth [i..\<le>j] k) = i + k"
using assms 
by (induct j arbitrary: k i, simp, 
    metis InfiniteInterval.upt.upt_Suc Nat.le_diff_conv2 add.commute add_leD1 
          interval_intapp_nth le_SucE upt_length upt_nth_help)

lemma upt_intfirst:
assumes "i \<le> j"
shows   "  intfirst [i..\<le>j] = i"
using assms
by (simp add: InfiniteInterval.upt_rec)

lemma upt_intlast:
 " intlast [i..\<le>j] = j"
by (metis InfiniteInterval.upt_rec interval_nth_intlen_intlast intlast.simps(1) less_or_eq_imp_le 
          ordered_cancel_comm_monoid_diff_class.add_diff_inverse upt_length upt_nth)

lemma prefix_upt:
assumes "i+m \<le> n"
shows   "  prefix m [i..\<le>n] = [i..\<le>i+m]"
using assms 
proof
 (induct m arbitrary: i)
 case 0
 then show ?case by (simp add: upt_intfirst upt_same)
 next
 case (Suc m)
 then show ?case using InfiniteInterval.upt_rec by auto
qed

lemma suffix_upt:
 " suffix m [i..\<le>j] = [i+m..\<le>j]"
proof
 (induct m arbitrary: i j)
 case 0
 then show ?case by simp
 next
 case (Suc j)
 then show ?case using InfiniteInterval.upt_rec 
 by (metis add_Suc_shift interval_suffix_suc not_less_eq not_less_iff_gr_or_eq suffix.simps(1))
qed

lemma map_suc_upt:
 " map Suc [m..\<le>n] = [Suc m..\<le>Suc n]"
proof
 (induct n arbitrary: m)
 case 0
 then show ?case by simp
 next
 case (Suc n)
 then show ?case by simp
qed

lemma map_add_upt:
 "map (\<lambda>i. i + n) [0..\<le>m] = [n..\<le>m+n]"
proof
 (induct m)
 case 0
 then show ?case by (simp add: upt_same)
 next
 case (Suc m)
 then show ?case by simp
qed

subsubsection \<open>isuffix\<close>

lemma isuffix_nth:
 "(isuffix k x) n = x (k+n)"
by (simp add: isuffix_def)

lemma isuffix_0:
 "isuffix 0 x = x"
by (simp add: isuffix_def)

lemma isuffix_isuffix:
 "(isuffix m (isuffix n x)) = isuffix (n+m) x"
by (rule ext) (simp add: isuffix_def add.assoc)

subsubsection \<open>iprefix\<close>

lemma iprefix_0:
 "(iprefix 0 x) = \<langle>(x 0)\<rangle>"
by (simp add: iprefix_def subinterval_def)

lemma iprefix_nth:
assumes "k\<le>m"
shows   "    (nth (iprefix m x) k) = (x k)"
using assms
by (simp add: interval_nth_map iprefix_def subinterval_def upt_nth)

lemma iprefix_length:
 "intlen (iprefix n x) = n"
by (simp add:  iprefix_def subinterval_def upt_length)


subsubsection \<open>subinterval\<close>

lemma subinterval_length:
 "intlen (subinterval x i j)  = j -i"
by (simp add: subinterval_def upt_length)

lemma subinterval_nth:
assumes "i+k \<le> j"
shows   "  nth (subinterval x i j) k = x (i+k)"
unfolding subinterval_def
using assms by (simp add: interval_nth_map upt_nth)

lemma iprefix_isuffix:
 "iprefix n (isuffix k x) = subinterval x k (n+k) "
proof -
 have 0: "iprefix n (isuffix k x) = map (\<lambda>n. x (k + n)) [0..\<le>n]"
   by (simp add: iprefix_def isuffix_def subinterval_def)
 have 1: " [k..\<le>n+k] =  (map (\<lambda>i. i+k) [0..\<le>n]) "
 using map_add_upt by simp
 hence 2: "map x [k..\<le>n+k] =  map x (map (\<lambda>i. i+k) [0..\<le>n])"
 by simp
 have 3: "map x (map (\<lambda>i. i+k) [0..\<le>n]) = map (x \<circ> (\<lambda>i. i+k)) [0..\<le>n] "
 by simp
 have 4: "(x \<circ> (\<lambda>i. i+k)) = (\<lambda>n. x (k + n)) " by (metis add.commute comp_apply)
 hence 5: "map (x \<circ> (\<lambda>i. i+k)) [0..\<le>n] = map (\<lambda>n. x (k + n)) [0..\<le>n] "
 by simp
 have 6: " subinterval x k (n+k) = map x [k..\<le>n+k] "
  by (simp add: subinterval_def)
 from 0 2 3 5 6 show ?thesis by auto
qed

lemma subinterval_sub_isuffix:
 assumes "i < j"
 shows   " (subinterval xs (i+k) (j+k)) = (subinterval (isuffix k xs) i j)"
proof -
 have 1: " (subinterval xs (i+k) (j+k)) = 
           iprefix (j-i) (isuffix (i+k) xs) "
  by (simp add: iprefix_isuffix assms less_imp_le_nat)
 have 2: "iprefix (j-i) (isuffix (i+k) xs) =
          iprefix (j-i) (isuffix (i) (isuffix k xs))"
 by (simp add: isuffix_isuffix add.commute)
 have 3: " iprefix (j-i) (isuffix (i) (isuffix k xs)) =
           (subinterval (isuffix k xs) i j)"
   by (simp add: iprefix_isuffix assms less_imp_le_nat)
 from 1 2 3 show ?thesis by auto
qed           

lemma subinterval_sub_isuffix_iidx:
assumes "infinite_index_sequence 0 lsk \<and> n>0"
shows   " (subinterval \<sigma> ((lsk i) +n) ((lsk (Suc i))+n)) =
   (subinterval (isuffix n \<sigma>) (lsk i) (lsk (Suc i)))"
using assms by (simp add: infinite_index_sequence_def subinterval_sub_isuffix)

lemma interval_pref_ipref_3_intlen:
 " intlen (prefix i (iprefix (i+k) xs)) = intlen (iprefix i xs)"
by (simp add: iprefix_length)

lemma interval_pref_ipref_3_nth:
 " (nth (prefix i (iprefix (i+k) xs)) m) = (nth (iprefix i xs) m)"
by (smt interval_eq_nth_eq interval_nth_prefix interval_pref_ipref_3_intlen iprefix_length 
        iprefix_nth le_add1 le_trans) 


lemma interval_pref_ipref_3 [simp]:
    " (prefix i (iprefix (i+k) xs)) = iprefix i xs"
by (simp add: interval_eq_nth_eq interval_pref_ipref_3_intlen interval_pref_ipref_3_nth)

lemma interval_iprefix_isuffix_swap_intlen:
 "intlen (iprefix ia (isuffix i xs)) = intlen (suffix i (iprefix (ia+i) xs))"
by (simp add: iprefix_length)

lemma interval_iprefix_isuffix_swap_nth:
 assumes "m\<le>ia"
  shows  " (nth (iprefix ia (isuffix i xs)) m) = (nth (suffix i (iprefix (ia+i) xs)) m)"
using assms by (simp add: iprefix_length iprefix_nth isuffix_def)

lemma interval_iprefix_isuffix_swap:
 " iprefix ia (isuffix i xs) = suffix i (iprefix (ia+i) xs)"
 by (simp add: interval_eq_nth_eq interval_iprefix_isuffix_swap_nth iprefix_length)


subsubsection \<open>Conc\<close>

lemma conc_empty_zero:
 "(conc \<langle>s\<rangle> x) 0 = s  "
unfolding conc_def by auto

lemma conc_empty_suc:
 "(conc \<langle>s\<rangle> x) (Suc i) = x i"
unfolding conc_def by auto

lemma conc_conc:
 "conc x (conc y w) = conc (x \<ominus> y) w" (is "?lhs = ?rhs")
proof
 fix n
 have x: "n \<le> intlen x \<Longrightarrow> ?lhs n = ?rhs n"
   by (simp add: conc_def interval_intapp_nth)
 have y: "n > intlen x \<and> n \<le> (intlen x) + (intlen y) \<Longrightarrow> ?lhs n = ?rhs n"
   by (simp add: conc_def interval_intapp_nth, arith)
 have w: "n > (intlen x) + (intlen y) \<Longrightarrow> ?lhs n = ?rhs n"
   by (simp add: conc_def interval_intapp_nth, arith)
 from x y w show "?lhs n = ?rhs n" using not_less by blast
qed

lemma conc_iprefix_isuffix:
 " x = conc (iprefix n x) (isuffix (Suc n) x)"
by (rule ext) 
   (induct n, simp add: conc_def iprefix_0 isuffix_def,
              simp add: conc_def iprefix_length iprefix_nth ,
              metis isuffix_def nat_le_linear not_less_eq_eq 
              ordered_cancel_comm_monoid_diff_class.add_diff_inverse)



subsubsection \<open>Infinite index sequence\<close>

lemma iidx_1:
 " l = (conc \<langle>(l 0)\<rangle> (\<lambda>x. (l (x+1)))) 
  " (is "?lhs = ?rhs")
proof
 fix n
 have x: "n \<le> intlen \<langle>(l 0)\<rangle> \<Longrightarrow> ?lhs n = ?rhs n"
 by (simp add: conc_def interval_intapp_nth)
 have ls: "n > intlen \<langle>(l 0)\<rangle> \<Longrightarrow> ?lhs n = ?rhs n"
 by (simp add: conc_def interval_intapp_nth)
 from x ls show "?lhs n = ?rhs n" by auto
qed

lemma iidx_2:
 "infinite_index_sequence 0 (conc \<langle>(l 0)\<rangle> (\<lambda>x. (l (x+1)))) =
  ( (l 0) = 0 \<and> (l 0) < (l 1) \<and> 
  infinite_index_sequence (l 1) (\<lambda>x. (l (x+1))))"
by (simp add: infinite_index_sequence_def conc_def, auto)
   (metis Suc_diff_Suc add_diff_cancel_left' gr0I plus_1_eq_Suc zero_less_diff)

lemma iidx_less_plus:
 assumes " infinite_index_sequence n ls"
 shows   " (ls i) < (ls (Suc (i+k))) "
using assms 
by (simp add: infinite_index_sequence_def lift_Suc_mono_less)

lemma iidx_greater:
 assumes "infinite_index_sequence n ls"
 shows   " i>0 \<longrightarrow> n < ls i"
using assms 
by (induct i, simp, metis infinite_index_sequence_def less_imp_Suc_add iidx_less_plus) 

lemma iidx_3:
 assumes "infinite_index_sequence n ls"
 shows   " infinite_index_sequence 0 ((shiftm n)\<circ>ls)  "
using assms 
by (simp add: infinite_index_sequence_def shiftm_def)
   (metis diff_less_mono less_le lift_Suc_mono_less_iff not_less zero_less_Suc)


lemma iidx_4:
 " (infinite_index_sequence (x+k) ((shift k)\<circ>ls))=
    infinite_index_sequence x ls "
by (simp add: shift_def infinite_index_sequence_def)


lemma iidx_5:
 assumes  " (infinite_index_sequence k (lsk) \<and> ls = (shiftm k)\<circ>lsk)"
 shows    "infinite_index_sequence 0 ls"
using assms 
by (simp add: infinite_index_sequence_def shiftm_def)    
   (metis add_less_same_cancel1 diff_less_mono lift_Suc_mono_less_iff not_add_less1 not_le_imp_less)

lemma iidx_ext:
 " ((xs :: infiniteindex ) = ys) = (\<forall>i. xs i = ys i)"
by auto

lemma iidx_6:
 "(infinite_index_sequence k lsk \<and> lsk = (shift k)\<circ>ls \<and> infinite_index_sequence 0 ls ) =
  (infinite_index_sequence k lsk \<and> ls = (shiftm k)\<circ>lsk \<and> infinite_index_sequence 0 ls)"
by (simp add: iidx_ext infinite_index_sequence_def shift_def shiftm_def,rule, auto,
  metis add.commute add_diff_inverse_nat add_less_same_cancel1 lift_Suc_mono_less_iff not_add_less1)


lemma iidx_7:
 " (infinite_index_sequence k lsk \<and> ls= (shiftm k) \<circ> lsk) =
   (infinite_index_sequence k lsk \<and> ls= (shiftm k) \<circ> lsk \<and>
    infinite_index_sequence 0 ls)"
using iidx_5 by blast


lemma iidx_8:
 " ( lsk = (shift k) \<circ> ls \<and> infinite_index_sequence 0 ls) =
   ( lsk = (shift k) \<circ> ls \<and> infinite_index_sequence k lsk \<and>
     infinite_index_sequence 0 ls )"
by (metis Interval.shift_def add_left_imp_eq diff_is_0_eq' interval_idx_shift_mono 
          le_add_diff_inverse2 mono_def iidx_4 rel_simps(46))

lemma iidx_0_a:
 " infinite_index_sequence 0 l \<Longrightarrow>
   (l 0 ) = 0 \<and> (l 0) < (l 1) \<and> infinite_index_sequence (l 1) (\<lambda>x. l(x+1))"
by (simp add: infinite_index_sequence_def)
    metis

lemma iidx_0_b:
 " x =0 \<and> x < (l 0) \<and> infinite_index_sequence (l 0) l \<Longrightarrow>
   infinite_index_sequence 0 (conc \<langle>x\<rangle> l)"
using diff_Suc_less infinite_index_sequence_def  
by (simp add: conc_def lift_Suc_mono_less)

lemma iidx_0:
 "(\<exists> l. infinite_index_sequence 0 l) =
  (\<exists> ls x. x =0 \<and> x< (ls 0) \<and> infinite_index_sequence (ls 0) ls)"
using infinite_index_sequence_def lessI iidx_0_b by metis


end