(*  title     : An encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)
section \<open>Finite Intervals\<close>

theory Interval
 imports
   Main 
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="cvc4 z3 e spass vampire " ]
*)
(*
declare [[smt_timeout=120]]
*)


text \<open>
   An interval is a finite sequence of elements of a particular type. Intervals are similar to list 
   in Isabelle/HOL, the difference is that intervals have instead of nil a single
   element at the end. So an interval of length zero is a single element whereas for Isabelle's list 
   we have that an empty list is nil (no element present). 

   The usual operations on intervals are defined: length (\<open>intlen\<close>), \<open>prefix\<close>, 
   \<open>suffix\<close>, \<open>sub\<close>,  \<open>nth\<close>, \<open>intfirst\<close>, \<open>intlast\<close>, \<open>intapp\<close> 
   and \<open>intrev\<close>. 

   We also introduce \<open>index_sequence\<close> which is a sequence of chop (fuse) points. This sequence
   is used in the old definition of chopstar which is an existential quantification over this 
   sequence. The type \<open>index_sequence\<close> is again of type interval but the elements are 
   natural numbers. Two functions \<open>shift\<close> and \<open>shiftm\<close> are introduced that are used to add (shift) 
   and subtract a natural number of each element in the sequence of chop (fuse) points.   
    
\<close>

subsection \<open>Definitions\<close>

datatype (set: 'a) interval =
    St  'a ("\<lceil>_\<rceil>")
  | Cons 'a "'a interval"  (infixr "\<odot>" 65)
for
  map: map
  rel: interval_all2
  pred: interval_all

(*print_theorems*)

type_synonym  index = "nat interval"  

syntax
  \<comment> \<open>interval Enumeration\<close>
  "_interval" :: "args => 'a interval"    ("\<langle>(_)\<rangle>")

translations
  "\<langle>x, xs\<rangle>" == "x\<odot>\<langle>xs\<rangle>"
  "\<langle>x\<rangle>" == "\<lceil>x\<rceil>"
  
primrec (nonexhaustive) intlen :: "'a interval \<Rightarrow> nat" where
    " intlen \<langle>x\<rangle> = 0"
  | " intlen (x\<odot>xs) = 1+ (intlen xs)"
 
primrec (nonexhaustive) nth :: "'a interval => nat => 'a" where
    "nth \<langle>x\<rangle> n      = x"
  | "nth (x\<odot>xs) n = (case n of 0 \<Rightarrow> x | Suc k \<Rightarrow> nth xs k)"
   
primrec prefix:: "nat \<Rightarrow> 'a interval \<Rightarrow> 'a interval" where
    "prefix n \<langle>x\<rangle> = \<langle>x\<rangle>"   
  | "prefix n (x \<odot>xs) = (case n of 0 \<Rightarrow> \<langle>x\<rangle> | Suc m \<Rightarrow> x \<odot> (prefix m xs))"

primrec suffix:: "nat \<Rightarrow> 'a interval \<Rightarrow> 'a interval" where
   "suffix n \<langle>x\<rangle> = \<langle>x\<rangle>" 
 | "suffix n (x\<odot>xs) = (case n of 0 \<Rightarrow> (x\<odot>xs)  | Suc m \<Rightarrow> suffix m xs)"

definition sub:: "nat \<Rightarrow> nat \<Rightarrow> 'a interval \<Rightarrow> 'a interval" 
where
 "sub n k xs = (if k<n then prefix 0 (suffix n xs)
                       else prefix (k-n) (suffix n xs)
               )" 

primrec intfirst :: "'a interval \<Rightarrow> 'a" where
    "intfirst \<langle>x\<rangle>     = x"
  | "intfirst (Cons x _) = x"

primrec intlast :: "'a interval \<Rightarrow> 'a" where
    "intlast \<langle>x\<rangle>      = x"
  | "intlast (Cons _ xs) = intlast xs"  

primrec intapp :: "'a interval \<Rightarrow> 'a interval \<Rightarrow> 'a interval" (infixr "\<ominus>" 65) where
intapp_St: "\<langle>x\<rangle> \<ominus> ys = x \<odot> ys" |
intapp_Cons: "(x\<odot>xs) \<ominus> ys = x \<odot> (xs \<ominus> ys)"

primrec intrev :: "'a interval \<Rightarrow> 'a interval" where
  " intrev \<langle>x\<rangle> = \<langle>x\<rangle>"
| " intrev (x \<odot>xs) = (intrev xs) \<ominus> \<langle>x\<rangle>"

definition index_sequence :: " nat \<Rightarrow> index \<Rightarrow> bool" where
  "index_sequence x idx \<equiv> (nth idx 0 = x) \<and> (\<forall> n. n<intlen idx \<longrightarrow> nth idx n < nth idx (Suc n))"  

definition shift :: "nat \<Rightarrow> nat \<Rightarrow> nat" where
  "shift k = (\<lambda> x. x+k)"

definition shiftm ::  "nat \<Rightarrow> nat \<Rightarrow> nat" where
  "shiftm k = (\<lambda> x. (if k>x then 0 else  (x-k)))" 



subsection \<open>Lemmas\<close>

text \<open>
 Basic lemmas are introduced for each of the above operations on intervals.  
\<close>

subsubsection \<open>Interval Length\<close>

lemma interval_intlen_gr_zero [simp]:
    "intlen xs \<ge> 0"
by auto

lemma interval_intlen_cons [simp]: 
    " (intlen (x\<odot>xs)) = (intlen xs) +1"
by simp
  
lemma interval_intlen_cons_1 :
    " intlen l > 0 \<longleftrightarrow> (\<exists> x ls. l = x \<odot>ls) "
by (induct l) simp_all

lemma interval_intlen_map [simp]:
    " intlen (map f xs) = intlen xs"
by (induct xs) simp_all

lemma interval_intlen_intapp [simp]:
 " intlen (xs \<ominus> ys) = (intlen xs) + (intlen ys) +1"
by (induct xs arbitrary: ys) simp_all

lemma interval_intrev_intlen [simp]:
 " intlen (intrev xs) = intlen xs"
by (induct xs, simp, simp)

subsubsection \<open>Map\<close>

lemma map_ext:
 "(\<And>x. x: set xs \<longrightarrow> f x = g x) \<Longrightarrow> map f xs = map g xs"
by (induct xs) simp_all

lemma map_ident [simp]:
 "map (\<lambda>x. x) = (\<lambda>xs .xs)"
by (rule ext, induct_tac xs) auto

lemma map_intapp [simp]:
 "map f (xs \<ominus> ys) = map f xs \<ominus> map f ys"
by (induct xs) auto

lemma map_map [simp]:
 "map f (map g xs) = map (f \<circ> g) xs"
by (simp add: interval.map_comp)

lemma map_comp_map [simp]:
 "((map f) \<circ> (map g)) = map(f \<circ> g)"
by (rule ext) simp

lemma intrev_map:
 "intrev (map f xs) = map f(intrev xs)"
by (induct xs) auto

lemma map_eq_conv [simp]:
 "(map f xs = map g xs) = (\<forall> x\<in> set xs. (f x) = (g x))"
by (induct xs) auto

lemma map_cong:
 "xs = ys \<Longrightarrow> (\<And>x. x \<in> set ys \<Longrightarrow> f x = g x) \<Longrightarrow> map f xs = map g ys"
by simp

lemma map_injective:
 "map f xs = map f ys \<Longrightarrow> inj f \<Longrightarrow> xs = ys"
by (meson injD interval.inj_map)

lemma inj_map_eq_map [simp]:
 "inj f \<Longrightarrow> (map f xs = map f ys)  = (xs = ys)"
by (blast dest: map_injective)

lemma inj_mapI:
 "inj f \<Longrightarrow> inj (map f)"
using interval.inj_map by blast

lemma inj_mapD:
 "inj (map f) \<Longrightarrow> inj f"
by (metis inj_def interval.map(1) interval.simps(1))

lemma inj_map[iff]:
 "inj (map f) = inj f"
by (blast dest: inj_mapD intro: inj_mapI)

lemma map_idI: "(\<And>x. x \<in> set xs \<Longrightarrow> f x = x) \<Longrightarrow> map f xs = xs"
by (induct xs, auto)

functor map: map
by (simp_all add: id_def)

declare map.id [simp]


subsubsection \<open>nth\<close>

lemma interval_nth_zero [simp]:
    " nth (x\<odot>xs) 0  = x"
by simp

lemma interval_nth_Suc [simp]:
    "nth (x\<odot>xs) (Suc n) = nth xs n"
by auto

lemma interval_nth_last:
    " nth (x\<odot>xs) (intlen (x\<odot>xs)) = nth xs (intlen xs) "
by simp

lemma interval_nth_cons:
 assumes "0<i \<and> i<1+intlen(xs)"
 shows   "nth(x\<odot>xs) i =nth xs (i-1) \<and>
          nth(x\<odot>xs) (i+1) = nth xs ((i-1)+1)"
by (metis One_nat_def Suc_leI add.commute assms interval_nth_Suc le_add_diff_inverse2 plus_1_eq_Suc)

lemma interval_nth_zero_intfirst:
 " nth xs 0 = intfirst xs "
by (induct xs) simp_all

lemma interval_nth_intlen_intlast:
 " nth xs (intlen xs) = intlast xs"
by (induct xs) simp_all

lemma interval_st_intlen :
    " (xs = \<langle>x\<rangle>) \<longleftrightarrow> intlen xs = 0 \<and> nth xs 0 = x" 
by (induct xs) simp_all

lemma interval_eq_nth_eq :
    " (xs =ys) = (intlen xs =intlen ys \<and> (\<forall> i\<le> intlen xs. nth xs i = nth ys i))"
by (induct xs arbitrary: ys, metis interval_st_intlen le_numeral_extra(3), 
    (case_tac ys, simp, fastforce))


lemma interval_nth_map :
    "nth (map f xs) i = f (nth xs i)"
by (induct xs arbitrary: i, simp, (case_tac i, simp, simp))


subsubsection \<open>index sequence\<close>

lemma interval_idx_less:
    assumes iseq: " index_sequence x idx"
    shows " (n<intlen idx \<and> n+k<intlen idx) \<longrightarrow> nth idx n < nth idx (Suc(n+k)) "
using index_sequence_def iseq by (induct k, simp, auto)

lemma interval_idx_less_last :
 assumes " index_sequence x idx"
 shows   " (i<intlen idx \<and> i+(intlen idx - (i+1))< intlen idx)
            \<longrightarrow> nth idx i < nth idx (Suc(i+(intlen idx -(i+1))))"
using assms interval_idx_less by blast

lemma interval_idx_less_last_1:
 assumes  "index_sequence x idx"
 shows    "i<intlen idx \<longrightarrow> nth idx i < nth idx (intlen idx) "
using assms interval_idx_less_last by auto

lemma interval_idx_greater_first:
 assumes  " index_sequence x idx"
 shows    " (i>0 \<and> i\<le> intlen idx) \<longrightarrow> x < nth idx i"
using assms by (induct i, simp, 
   metis One_nat_def Suc_le_lessD add_Suc index_sequence_def interval_idx_less 
          less_le_trans plus_1_eq_Suc)

lemma interval_idx_cons:
    " index_sequence 0 (x\<odot>ls) = 
       (x=0 \<and> x<nth ls 0 \<and> index_sequence (nth ls 0) ls)"
using less_Suc_eq_0_disj by (simp add: index_sequence_def, auto)

lemma interval_idx_shift_mono:
    "mono (shift k)" 
by (simp add: Interval.shift_def mono_def)     

lemma interval_idx_expand:
 assumes  "index_sequence 0 l  \<and> (nth l (intlen l)) = (intlen xs) \<and> 0\<le>i \<and> i< (intlen l)"
 shows    " 0\<le> (nth l i) \<and> (nth l i) \<le>  (nth l (i+1)) \<and> (nth l (i+1)) \<le> (intlen xs)"
using assms
by (simp add: index_sequence_def, 
    (induct l, simp, metis Suc_lessI eq_imp_le interval_idx_less_last_1 less_imp_le_nat))

lemma interval_idx_shift_idx [simp]:
    " ( index_sequence (x+k) (map (shift k)  idx)) = (index_sequence x idx) " 
by (simp add: Interval.shift_def index_sequence_def interval_nth_map)    

lemma interval_idx_shiftm :
  assumes "(index_sequence k (lsk) \<and> ls = map (shiftm k) lsk)"
 shows    "index_sequence 0 (ls) \<and> (intlen ls) = (intlen lsk)"
using assms
by (simp add: interval_eq_nth_eq index_sequence_def shift_def shiftm_def interval_nth_map ) 
   (smt Suc_leI diff_less_mono index_sequence_def interval_idx_greater_first interval_intlen_map 
        le_less_trans less_Suc_eq_0_disj not_less order.asym)

lemma interval_lsk_ls :
    "(index_sequence k (lsk) \<and> lsk = map (shift k) ls \<and> index_sequence 0 (ls) ) =
      (index_sequence k (lsk) \<and> ls = map (shiftm k) lsk \<and> index_sequence 0 (ls) ) "
by (simp add: interval_eq_nth_eq index_sequence_def shift_def shiftm_def interval_nth_map,
    rule,
    metis (no_types, lifting) add_diff_cancel_right' not_add_less2,
    metis (no_types, lifting) Suc_eq_plus1 add.commute add_cancel_right_left add_diff_inverse_nat 
       ex_least_nat_less le_SucE le_zero_eq not_less_zero order_refl)

lemma interval_idx_link_shiftm:
    "(index_sequence k (lsk) \<and> ls = map (shiftm k) lsk ) =
     (index_sequence k (lsk) \<and> ls = map (shiftm k) lsk \<and> 
      index_sequence 0 (ls) \<and> (intlen ls) =(intlen lsk)) "
using interval_idx_shiftm by blast
    
lemma interval_idx_link:
    "(lsk = map (shift k) ls \<and> index_sequence 0 (ls) ) =
     (lsk = map (shift k) ls \<and> index_sequence k (lsk) \<and> index_sequence 0 (ls)\<and> 
      (intlen ls) =(intlen lsk))"
by (metis add.left_neutral interval_idx_shift_idx interval_intlen_map)

lemma interval_idx_bound_0 :
 assumes "index_sequence 0 ls \<and> Interval.nth ls (intlen ls) = intlen (suffix k xs)"
 shows   "((i\<le>intlen ls) \<longrightarrow> ((nth ls (i)) \<le> (intlen (suffix k xs))))"
using assms
by (metis eq_iff interval_idx_less_last_1 le_neq_implies_less less_imp_le_nat) 
 
lemma interval_idx_bound_1:
  "(index_sequence 0 (ls) \<and> (nth (ls) (intlen (ls))) = (intlen (suffix k xs))) =
     (index_sequence 0 (ls) \<and> (nth (ls) (intlen (ls))) = (intlen (suffix k xs)) \<and>
     (\<forall>i. (i\<le>intlen ls) \<longrightarrow> ((nth ls (i)) \<le> (intlen (suffix k xs)))) )" 
using interval_idx_bound_0 by blast

subsubsection \<open>prefix, suffix and sub\<close>

lemma interval_prefix_state [simp]: 
    " prefix m \<langle>x\<rangle> = \<langle>x\<rangle> " 
by simp

lemma interval_prefix_suc [simp]:
    " prefix (Suc m) (x\<odot>xs) = x \<odot> (prefix m xs)" 
by auto

lemma interval_prefix_zero [simp]:
    " prefix 0 (x\<odot>xs) = \<langle>x\<rangle>"
by auto

lemma interval_prefix_zero_intfirst [simp]:
  " prefix 0 xs = \<langle>intfirst xs\<rangle>"
by (induct xs) simp_all

lemma interval_intfirst_prefix [simp]:
 " i\<le>intlen xs \<Longrightarrow>   intfirst (prefix i xs) = intfirst xs"
by (induct xs arbitrary: i, auto) (case_tac i, auto)
  
lemma interval_prefix_intlen [simp]:
    " (prefix (intlen xs) xs) = xs"
by (induct xs) simp_all

lemma interval_prefix_intlen_gr_1 [simp]: 
    "(prefix ((intlen xs)+i) xs) = xs"
by (induct xs) simp_all

lemma interval_intlen_prefix_cons [simp]:
    " intlen( prefix (Suc i) (x\<odot>xs)) = 1 + intlen(prefix i xs)"
using interval_intlen_cons by auto

lemma interval_prefix_length [code]: 
    " intlen (prefix i xs) = (if i\<le> intlen xs then i else intlen xs)"
by (induct xs arbitrary: i, simp) (case_tac i, auto) 

lemma interval_prefix_length_good [simp]:
   assumes " i\<le> intlen xs"
   shows   " (intlen ( prefix i xs)) = i" 
using assms by (simp add: interval_prefix_length)

lemma interval_prefix_length_bad [simp] : 
  assumes   " i > intlen xs"
  shows     "  intlen (prefix i xs) = intlen xs"
using assms by (simp add: interval_prefix_length)
   
lemma interval_pref_intlen_bound :
 assumes   " i \<le> (intlen xs)"
 shows     "  intlen (prefix i xs) \<le> intlen xs " 
using assms by (induct xs, simp) (metis interval_prefix_length)

lemma interval_suffix_length [code]:
    " intlen (suffix i xs ) = (if i\<le> intlen xs then (intlen xs)-i else 0) "
by (induct xs arbitrary: i, simp) (case_tac i, auto)
    
lemma interval_suffix_length_good [simp]:
   assumes " i\<le> intlen xs"
   shows   " intlen (suffix i xs ) = (intlen xs)-i"
using assms by (simp add: interval_suffix_length)
  
lemma interval_suffix_length_bad [simp]:
  assumes  " i> intlen xs"
  shows    "  intlen (suffix i xs ) = 0"
using assms by (simp add: interval_suffix_length)
 
lemma interval_nth_prefix [simp]:
 " i\<le>intlen xs \<and> k \<le> i \<Longrightarrow> nth (prefix i xs) k = nth xs k"
by (induct xs arbitrary: i k, auto, (case_tac i, auto, (case_tac k, auto)))

lemma interval_nth_suffix [simp]:
 assumes " i\<le>intlen xs \<and> k \<le> intlen xs -i"
 shows   " nth (suffix i xs) k = nth xs (i+k)"
by (induct xs arbitrary: i k, auto) (case_tac i, auto)

lemma interval_suffix_prefix_help_1:
 assumes " ia+i \<le> intlen xs \<and> k \<le> ia"
 shows " nth (prefix ia (suffix i xs)) k = nth (suffix i (prefix (ia +i) xs)) k"
proof -
 have 1: " nth (prefix ia (suffix i xs)) k = nth (suffix i xs) k" 
 using interval_nth_prefix assms by (metis interval_prefix_intlen_gr_1 le_cases le_iff_add) 
 have 2: " nth (suffix i xs) k = nth xs (i+k) " 
 using interval_nth_suffix assms by (simp add: add_le_imp_le_diff)
 have 3: " nth xs (i+k) = nth (prefix (ia+i) xs) (i+k)" 
 using interval_nth_prefix assms by simp
 have 4: " nth (prefix (ia+i) xs) (i+k) = nth (suffix i (prefix (ia +i) xs)) k" 
 using interval_nth_suffix assms by simp
 from 1 2 3 4 show ?thesis by auto
qed

lemma interval_suffix_prefix_help_2:
 assumes "ia+i \<le> intlen xs"
 shows   " (\<forall> k \<le> ia . nth (prefix ia (suffix i xs)) k = nth (suffix i (prefix (ia +i) xs)) k)"
using interval_suffix_prefix_help_1 using assms by fastforce

lemma interval_suffix_prefix_help_3:
 assumes "ia+i \<le> intlen xs "
 shows   "intlen (prefix ia (suffix i xs)) = intlen (suffix i (prefix (ia +i) xs)) "
using assms interval_prefix_length_good interval_suffix_length_good by auto

lemma interval_suffix_prefix_swap:
 assumes "ia+i \<le> intlen xs"
 shows   " prefix ia (suffix i xs) = suffix i (prefix (ia +i) xs)"
by (simp add: interval_eq_nth_eq interval_suffix_prefix_help_2 interval_suffix_prefix_help_3 assms)
 
lemma interval_prefix_prefix_zero [simp]:
    " prefix 0 ( prefix 0 xs ) = prefix 0 xs" 
by (induct xs) simp_all
  
lemma interval_pref_pref [simp]:
    " (prefix i (prefix i xs)) = prefix i xs"
by (metis interval_prefix_intlen interval_prefix_intlen_gr_1 interval_prefix_length 
          less_imp_add_positive not_less)

lemma interval_pref_pref_3 [simp]:
    " (prefix i (prefix (i+k) xs)) = prefix i xs"
by (induct xs arbitrary: i k, simp, (case_tac i, auto, (simp add: Nitpick.case_nat_unfold)))
 
lemma interval_pref_help:
 assumes "i\<le>intlen (prefix (intlen xs - Suc 0) xs)"
 shows   " (prefix i (prefix (intlen xs - Suc 0) xs)) = (prefix i xs)  "
using assms 
by (metis diff_le_self interval_pref_pref_3 interval_prefix_length 
           ordered_cancel_comm_monoid_diff_class.add_diff_inverse)

lemma interval_pref_pref_help:
 assumes "intlen xs >0 \<and> ia<intlen (xs)"
 shows   "(prefix ia (prefix (intlen xs - Suc 0) xs)) = (prefix ia  xs) "
using assms 
by (metis Suc_leI Suc_le_mono Suc_pred diff_le_self interval_pref_help interval_prefix_length_good)

lemma interval_pref_pref_help_1:
 assumes "i>0 \<and> i\<le> intlen xs"
 shows   "(prefix (intlen (prefix i xs) - Suc 0) (prefix i xs)) = 
          (prefix (intlen (prefix i xs) - Suc 0) xs) "
using assms interval_pref_pref_3 by (metis diff_le_self interval_prefix_length_good le_iff_add)

lemma interval_suffix_suc [simp]:
    " suffix (Suc m) (x \<odot> xs) = suffix m xs"
by auto
  
lemma interval_suffix_zero [simp]:
    " suffix 0 xs = xs "
by (induct xs) simp_all
    
lemma interval_suffix_intlen [simp]:
    " suffix (intlen xs) xs = \<langle>(nth xs (intlen xs))\<rangle> "
by (induct xs) simp_all

lemma interval_suffix_intlast [simp]:
   " suffix (intlen xs) xs = \<langle>intlast xs\<rangle>"
by (induct xs) simp_all

lemma interval_suffix_suffix [simp]:
    " suffix i (suffix j xs) = suffix (i+j) xs "
by (induct xs arbitrary: i j, simp, (case_tac i, auto, (simp add: Nitpick.case_nat_unfold)))

lemma interval_prefix_suffix_intlen [code]:
    "intlen (prefix ia (suffix i xs)) =
     (if i \<le> intlen xs then
      (if ia\<le> intlen xs -i  then ia else (intlen xs) -i )
      else 0)"
by (metis interval_prefix_length interval_suffix_length le_zero_eq)

lemma interval_prefix_suffix_intlen_good [simp]:
  assumes  "ia\<le> intlen xs -i \<and> i \<le> intlen xs"
  shows    " intlen (prefix ia (suffix i xs)) = ia "
using assms by (simp add: interval_prefix_suffix_intlen)

lemma interval_prefix_suffix_intlen_bad_0 [simp]:
  assumes  "i> intlen xs"
  shows    "intlen (prefix ia (suffix i xs)) = 0"
using assms by (simp add: interval_prefix_suffix_intlen)

lemma interval_prefix_suffix_intlen_bad_1 [simp] :
  assumes  "i \<le> intlen xs \<and> ia > intlen xs -i"
  shows    "intlen (prefix ia (suffix i xs)) = (intlen xs) -i  "
using assms by (simp add: interval_prefix_suffix_intlen)

lemma interval_suffix_suffix_3:
 assumes "i>0 \<and> ia<i \<and> i\<le> intlen xs"
 shows   "(suffix (i-ia) (suffix ((intlen xs)-i) xs)) = (suffix (((intlen xs)-ia)) xs)"
using assms by simp

lemma interval_sub_zero_prefix :
    " sub 0 k xs = prefix k xs"
by (simp add: Interval.sub_def) 

lemma interval_sub_suffix : 
  assumes  "(i < j \<and> j\<le> (intlen xs)-k)"
  shows    "(sub (i+k) (j+k) xs) = (sub i j (suffix k xs))"
using assms by (simp add: Interval.sub_def) 
  
lemma interval_sub_prefix_suffix_0:
  assumes "(0 \<le> i \<and> ia+i \<le> intlen xs)"
  shows   "  (sub i (i+ia) xs) = (prefix (ia) (suffix i xs))" 
using assms by (simp add: Interval.sub_def) 
  
lemma interval_sub_prefix_suffix:
 assumes " 0 \<le> i \<and> i\<le>j \<and> j \<le> intlen xs"
 shows   "  (sub i j xs) = (prefix (j-i) (suffix i xs))"
using assms by (simp add: Interval.sub_def) 



subsubsection \<open>intapp\<close>

lemma interval_prefix_intapp [simp]:
 "prefix (intlen xs -k) (xs \<ominus> ys) = prefix (intlen xs -k) xs"
by (induct xs arbitrary:k,simp, (case_tac k, simp, 
    metis comm_monoid_add_class.add_0 interval_intlen_gr_zero interval_prefix_intlen 
             le_add_diff_inverse),
    smt Suc_diff_Suc diff_Suc_Suc diff_is_0_eq' intapp.simps(2) interval_prefix_suc 
        interval_prefix_zero intlen.simps(2) not_le plus_1_eq_Suc)


lemma interval_prefix_intapp2 [simp]:
 "prefix (intlen xs + k +1) (xs \<ominus> ys) = xs \<ominus> prefix k ys"
by (induct xs arbitrary: k, simp, 
   (case_tac k, simp, metis add.right_neutral interval_prefix_zero_intfirst),
    simp add: add_eq_if)

lemma interval_suffix_intapp [simp]:
 "suffix (intlen xs +m +1) (xs \<ominus> ys) = suffix (m) ys"
by (induct xs arbitrary:m, simp, (case_tac m,simp,metis add.right_neutral interval_suffix_zero),
    simp add: add_eq_if)

lemma interval_suffix_intapp2 [simp]: 
 "(suffix (intlen xs - k)  xs) \<ominus> ys = suffix (intlen xs - k) (xs \<ominus> ys)"
by (induct xs, simp)
   (metis Suc_diff_le diff_is_0_eq' intapp_Cons interval_suffix_suc interval_suffix_zero 
          intlen.simps(2) not_less_eq_eq plus_1_eq_Suc)

lemma interval_intapp_assoc [simp]:
 " (xs \<ominus> ys) \<ominus> zs = xs \<ominus> ( ys \<ominus> zs ) "
by (induct xs) simp_all

lemma interval_intapp_nth:
 " nth (xs \<ominus> ys) k = (if k \<le> intlen xs 
                      then (nth xs k) 
                      else (nth ys (k - (intlen xs) -1)) )"
by (induct xs arbitrary: k, (case_tac k, simp, simp), 
    (case_tac k, simp, simp))

lemma interval_rev_intapp [simp]:
 " intrev (xs \<ominus> ys) = (intrev ys) \<ominus> (intrev xs)"
by (induct xs) simp_all

lemma interval_intlast_intapp [simp]:
 "intlast(xs \<ominus> \<langle>x\<rangle>) = x"
by (induct xs,simp,simp)

lemma interval_intlast_intapp2 [simp]:
 "intlast (xs \<ominus> ys) = intlast ys"
by (induct xs arbitrary: ys,simp,simp)

lemma interval_intfirst_intapp [simp]:
 "intfirst (\<langle>x\<rangle> \<ominus> xs) = x"
by (induct xs,simp,simp)

lemma interval_intfirst_intapp2 [simp]:
 "intfirst(xs \<ominus> ys) = intfirst xs"
by (induct xs arbitrary: ys, simp,simp)


subsubsection \<open>Reverse\<close>


lemma interval_rev_rev_ident [simp]: 
 " intrev (intrev xs) = xs"
by (induct xs) auto

lemma interval_rev_swap :
 " ((intrev xs) = ys) = (xs = intrev ys)"
by auto

lemma interval_rev_singleton_conv [simp]:
 "( intrev xs = \<langle>x\<rangle>) = (xs = \<langle>x\<rangle>)"
by (metis interval_rev_rev_ident intrev.simps(1)) 

lemma interval_single_rev_conv [simp]:
 " (\<langle>x\<rangle> = intrev xs) = (\<langle>x\<rangle> = xs)"
by (metis interval_rev_rev_ident intrev.simps(1))

lemma interval_rev_is_rev_conv [iff]:
 "(intrev xs = intrev ys) = (xs = ys)"
proof
 (induct xs arbitrary: ys)
 case (St x)
 then show ?case by simp
 next
 case (Cons x1a xs)
 then show ?case 
 using interval_rev_swap by force
qed

lemma interval_rev_induct [case_names St snoc]:
 "\<lbrakk>\<And> y. P \<langle>y\<rangle> ;\<And>x xs. P xs \<Longrightarrow> P(xs \<ominus> \<langle>x\<rangle>)\<rbrakk> \<Longrightarrow> P xs"
by (simplesubst interval_rev_rev_ident[symmetric],
  rule_tac interval = "intrev xs" in interval.induct, simp_all)

lemma interval_rev_exhaust [case_names St snoc]:
"(\<And>x. xs = \<langle>x\<rangle> \<Longrightarrow> P) \<Longrightarrow> (\<And>ys y. xs = ys \<ominus> \<langle>y\<rangle> \<Longrightarrow> P) \<Longrightarrow> P"
by (induct xs rule:interval_rev_induct) auto

lemmas interval_rev_cases = interval_rev_exhaust

lemma interval_rev_eq_cons_iff [iff]:
 "(intrev xs = y\<odot>ys) = (xs = (intrev ys) \<ominus> \<langle>y\<rangle>)"
by (metis interval_rev_rev_ident intrev.simps(2)) 

lemma interval_same_intapp_eq[iff]:
 "(xs \<ominus> ys = xs \<ominus> zs) = (ys = zs)"
using interval_suffix_intapp by (metis interval_suffix_zero)

lemma interval_intapp_eq_conv[iff]:
 "(xs \<ominus> \<langle>x\<rangle> = ys \<ominus> \<langle>y\<rangle>) = (xs = ys \<and> x = y)"
by (metis interval_intlast_intapp interval_rev_intapp interval_rev_is_rev_conv 
          interval_same_intapp_eq)

lemma interval_intapp_same_eq[iff]:
 "(ys \<ominus> xs = zs \<ominus> xs) = (ys = zs)"
by (metis interval_rev_intapp interval_rev_swap interval_same_intapp_eq)

lemma interval_intrev_intapp_cons:
 " intrev (xs \<ominus> \<langle>x\<rangle>) = x\<odot>intrev xs"
by (case_tac xs,simp,simp)

lemma interval_intlast_intrev:
 "intlast (intrev xs) = intfirst xs"
by (case_tac xs,simp,simp)

lemma interval_intfirst_intrev:
 "intfirst (intrev xs) = intlast xs"
by (induct xs,simp,simp)

lemma interval_intrev_nth:
assumes "k\<le> intlen (intrev xs)"
shows   "  (nth  (intrev xs) k) = (nth  xs ((intlen xs) -k))"
using assms 
by (induct xs, simp,simp,
   (case_tac k,simp add:interval_intapp_nth,
    smt Interval.nth.simps(1) Suc_diff_Suc diff_Suc_Suc diff_is_0_eq' interval_intapp_nth 
        interval_intrev_intlen le_SucE less_Suc_eq_le old.nat.simps(4) old.nat.simps(5)))


lemma interval_intrev_prefix:
assumes "k\<le> intlen xs"
shows   "  intrev( prefix k xs) = suffix ((intlen xs) - k) (intrev xs)  "
by (induct xs arbitrary: k, simp,simp,
    (case_tac k, simp,
     metis Suc_eq_plus1 add.right_neutral interval_intlast_intapp interval_intlen_intapp 
             interval_intrev_intlen interval_suffix_intlast intlen.simps(1)),
     metis  diff_Suc_Suc interval_intrev_intlen interval_suffix_intapp2 
          intrev.simps(2) old.nat.simps(5))

lemma interval_intrev_suffix:
assumes "k\<le> intlen xs"
shows   " intrev( suffix k xs) = prefix ((intlen xs) - k) (intrev xs)  "
using assms
by (induct xs arbitrary: k, simp, simp add: interval_intrev_prefix interval_rev_swap)

lemma interval_intrev_sub:
 assumes " 0 \<le> i \<and> i\<le>j \<and> j \<le> intlen xs"
 shows   " intrev (sub i j xs) = sub ((intlen xs) -j) ((intlen xs) - i) (intrev xs)"
using assms 
proof -
 have 1: "intrev (sub i j xs) = intrev (prefix (j-i) (suffix i xs))" 
   using assms interval_sub_prefix_suffix by (simp add: interval_sub_prefix_suffix)
 have 2: "intrev (prefix (j-i) (suffix i xs)) = suffix ((intlen xs) - j) (intrev (suffix i xs))"
   using assms interval_intrev_prefix[of "j-i" "suffix i xs"]  by auto
 have 3: "suffix ((intlen xs) - j) (intrev (suffix i xs)) = 
           suffix ((intlen xs) -j) (prefix ((intlen xs) - i) (intrev xs)) "
    using assms interval_intrev_suffix[of "i" "xs"] by auto
 have 4: "suffix ((intlen xs) -j) (prefix ((intlen xs) - i) (intrev xs)) =
          sub ((intlen xs) -j) ((intlen xs) - i) (intrev xs) "
    using assms by (simp add: diff_le_mono2 interval_sub_prefix_suffix interval_suffix_prefix_swap)
 from 1 2 3 4 show ?thesis by auto
qed




end
