(*  Title     : A Shallow Encoding of ITL in Isabelle/HOL
    Authors   : Antonio Cau     <cau.researcher at gmail.com>
                Ben Moszkowski
                David Smallwood <drs at dmu.ac.uk>
    Maintainer: Antonio Cau     <cau.researcher at gmail.com>        
    License   : BSD
*)

section \<open>Finite ITL Semantics\<close>   

theory Semantics
imports Interval "HOL-TLA.Intensional" 
begin
(*
sledgehammer_params [minimize=true,preplay_timeout=10,timeout=60,verbose=true,
                    provers="
  cvc4 z3 e spass vampire
"]
*)

text \<open> 
  This theory mechanises a \emph{shallow} embedding of finite ITL using the
  \<open>Interval\<close> and \<open>Intensional\<close> theories. A shallow embedding
  represents ITL using Isabelle/HOL predicates, while a \emph{deep}
  embedding \cite{ITLHOLv1} would represent ITL formulas  as mutually
  inductive datatypes. See, e.g., \cite{Wildmoser04} for a discussion
  about deep vs. shallow embeddings in Isabelle/HOL.  
  The choice of a shallow over a deep embedding is motivated \cite{Merz98,GrovMerz}
   by the following 
  factors: a shallow embedding is usually less involved, and existing Isabelle
  theories and tools can be applied more directly to enhance automation; due to
  the lifting in the \<open>Intensional\<close> theory, a shallow embedding can reuse
  standard logical operators, whilst a deep embedding requires a different
  set of operators for formulas. Finally, since our 
  target is system verification rather than proving meta-properties of the logic,
  which requires a deep embedding, a shallow embedding is more fit for purpose.
\<close>

subsection "Types of Formulas"

text \<open>
  To mechanise the ITL semantics, the following type abbreviations are used:
\<close>

type_synonym ('a,'b) formfun = "'a interval \<Rightarrow> 'b"
type_synonym 'a formula      = "('a,bool) formfun"
type_synonym ('a,'b) stfun   = "'a \<Rightarrow> 'b"
type_synonym 'a stpred       = "('a,bool) stfun"

instance 
 "fun" :: (type,type) world ..

instance
 "prod" :: (type,type) world ..

instance
 "interval" :: (type) world ..

text \<open>
  Pair, function, and interval are instantiated to be of type class world.
  This allows use of the lifted Intensional logic for formulas, and 
  standard logical connectives can therefore be used.
\<close>

subsection \<open>Semantics of ITL\<close>

text \<open> The semantics of ITL is defined. Note chopstar is a derived operator, i.e., it is 
defined recursively in terms of chop.\<close>

definition skip_d :: "('a ::world) formula "
where "skip_d \<equiv> \<lambda>s. intlen s =1"

definition chop_d :: " ('a ::world) formula \<Rightarrow> ('a ::world) formula \<Rightarrow> ('a ::world) formula"
where "chop_d F1 F2 \<equiv> \<lambda> s. \<exists>n. 0\<le> n \<and> n\<le>intlen s \<and> ((prefix n s) \<Turnstile> F1) \<and> ((suffix n s) \<Turnstile> F2 )"  

definition current_val_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun"
where "current_val_d f \<equiv> \<lambda> s. (nth s 0) \<Turnstile> f"

definition next_val_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun"
where "next_val_d f \<equiv> \<lambda> s. if intlen s >0 then ((nth s 1) \<Turnstile> f) else (\<some> (x::'b ). x=x) " 

definition fin_val_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun"
where "fin_val_d f \<equiv> \<lambda> s. (nth s (intlen s)) \<Turnstile> f"

definition penult_val_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun"
where "penult_val_d f \<equiv> \<lambda> s. if intlen s >0 then (nth s ((intlen s)-1) \<Turnstile> f) else (\<some> (x::'b ). x=x)" 


text \<open> This is the concrete syntax for the (abstract) operators above. \<close>

syntax
 "_skip_d"         :: "lift"                ("(skip)")
 "_chop_d"         :: "[lift,lift] \<Rightarrow> lift" ("(_;_)" [84,84] 83) 
 "_current_val_d"  :: "lift \<Rightarrow> lift"        ("($_)" [100] 99)
 "_next_val_d"     :: "lift \<Rightarrow> lift"        ("(_$)" [100] 99)
 "_fin_val_d"      :: "lift \<Rightarrow> lift"        ("(!_)" [100] 99)
 "_penult_val_d"   :: "lift \<Rightarrow> lift"        ("(_!)" [100] 99)
 "TEMP"            :: "lift \<Rightarrow> 'b"          ("(TEMP _)")

syntax (ASCII)
 "_skip_d"         :: "lift"                ("(skip)")
 "_chop_d"         :: "[lift,lift] \<Rightarrow> lift" ("(_;_)" [84,84] 83)
 "_current_val_d"  :: "lift \<Rightarrow> lift"        ("($_)" [100] 99)
 "_next_val_d"     :: "lift \<Rightarrow> lift"        ("(_$)" [100] 99)
 "_fin_val_d"      :: "lift \<Rightarrow> lift"        ("(!_)" [100] 99)
 "_penult_val_d"   :: "lift \<Rightarrow> lift"        ("(_!)" [100] 99)
translations
 "_skip_d"        \<rightleftharpoons> "CONST skip_d"
 "_chop_d"        \<rightleftharpoons> "CONST chop_d"
 "_current_val_d" \<rightleftharpoons> "CONST current_val_d"
 "_next_val_d"    \<rightleftharpoons> "CONST next_val_d"
 "_fin_val_d"     \<rightleftharpoons> "CONST fin_val_d"
 "_penult_val_d"  \<rightleftharpoons> "CONST penult_val_d" 
 "TEMP F"         \<rightharpoonup> "(F:: (_ interval) \<Rightarrow> _)"


subsection "Abbreviations"

text \<open> Some standard temporal abbreviations, with their concrete syntax. \<close>

definition sometimes_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "sometimes_d F \<equiv> LIFT(#True;F)"

definition di_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "di_d F \<equiv> LIFT(F;#True)" 

definition da_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "da_d F \<equiv> LIFT(#True;(F;#True))" 

definition next_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "next_d F \<equiv> LIFT(skip;F)" 

definition prev_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "prev_d F \<equiv> LIFT(F;skip)"

syntax
  "_sometimes_d" :: "lift \<Rightarrow> lift" ("(\<diamond>_)" [88] 87)
  "_di_d"        :: "lift \<Rightarrow> lift" ("(di _)" [88] 87)
  "_da_d"        :: "lift \<Rightarrow> lift" ("(da _)" [88] 87)
  "_next_d"      :: "lift \<Rightarrow> lift" ("(\<circle> _)" [88] 87)
  "_prev_d"      :: "lift \<Rightarrow> lift" ("(prev _)" [88] 87)


syntax (ASCII)
  "_sometimes_d" :: "lift \<Rightarrow> lift" ("(<>_)" [88] 87)
  "_di_d"        :: "lift \<Rightarrow> lift" ("(di _)" [88] 87) 
  "_da_d"        :: "lift \<Rightarrow> lift" ("(da _)" [88] 87)
  "_next_d"      :: "lift \<Rightarrow> lift" ("(next _)" [88] 87)
  "_prev_d"      :: "lift \<Rightarrow> lift" ("(prev _)" [88] 87)


translations 
  "_sometimes_d" \<rightleftharpoons> "CONST sometimes_d"
  "_di_d"        \<rightleftharpoons> "CONST di_d" 
  "_da_d"        \<rightleftharpoons> "CONST da_d"
  "_next_d"      \<rightleftharpoons> "CONST next_d" 
  "_prev_d"      \<rightleftharpoons> "CONST prev_d"


definition always_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "always_d F \<equiv> LIFT(\<not>(\<diamond>(\<not>F)))"

definition bi_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "bi_d F \<equiv> LIFT(\<not>(di(\<not>F)))"

definition ba_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "ba_d F \<equiv> LIFT(\<not>(da(\<not>F)))"

definition wnext_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "wnext_d F \<equiv> LIFT(\<not>(\<circle>(\<not>F)))"

definition wprev_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "wprev_d F \<equiv> LIFT(\<not>(prev(\<not>F)))"

definition more_d :: "('a::world) formula "
where "more_d \<equiv> LIFT(\<circle>(#True))" 


syntax
 "_always_d"   :: "lift \<Rightarrow> lift" ("(\<box>_)" [88] 87)
 "_bi_d"       :: "lift \<Rightarrow> lift" ("(bi _)" [88] 87)
 "_ba_d"       :: "lift \<Rightarrow> lift" ("(ba _)" [88] 87)
 "_wnext_d"    :: "lift \<Rightarrow> lift" ("(wnext _)" [88] 87)
 "_wprev_d"    :: "lift \<Rightarrow> lift" ("(wprev _)" [88] 87)
 "_more_d"     :: "lift"         ("(more)")
 

syntax (ASCII)
 "_always_d"   :: "lift \<Rightarrow> lift" ("([]_)" [88] 87)
 "_bi_d"       :: "lift \<Rightarrow> lift" ("(bi _)" [88] 87)
 "_ba_d"       :: "lift \<Rightarrow> lift" ("(ba _)" [88] 87)
 "_wnext_d"    :: "lift \<Rightarrow> lift" ("(wnext _)" [88] 87)
 "_wprev_d"    :: "lift \<Rightarrow> lift" ("(wprev _)" [88] 87)
 "_more_d"     :: "lift"         ("(more)")
 

translations
 "_always_d" \<rightleftharpoons> "CONST always_d"
 "_bi_d"     \<rightleftharpoons> "CONST bi_d" 
 "_ba_d"     \<rightleftharpoons> "CONST ba_d"
 "_wnext_d"  \<rightleftharpoons> "CONST wnext_d"
 "_wprev_d"  \<rightleftharpoons> "CONST wprev_d"
 "_more_d"   \<rightleftharpoons> "CONST more_d" 

definition empty_d :: "('a::world) formula "
where "empty_d \<equiv> LIFT(\<not>(more))" 

definition dm_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "dm_d F \<equiv> LIFT(#True;(more \<and> F))" 

syntax
 "_empty_d"    :: "lift"         ("(empty)")
 "_dm_d"       :: "lift \<Rightarrow> lift" ("(dm _)" [88] 87)

syntax (ASCII)
 "_empty_d"    :: "lift"         ("(empty)")
 "_dm_d"       :: "lift \<Rightarrow> lift" ("(dm _)" [88] 87)

translations
 "_empty_d" \<rightleftharpoons> "CONST empty_d"
 "_dm_d"    \<rightleftharpoons> "CONST dm_d"


definition bm_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "bm_d F \<equiv> LIFT(\<not>(dm(\<not>F)))" 

definition init_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "init_d F \<equiv> LIFT((empty \<and> F);#True)" 

definition fin_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "fin_d F \<equiv> LIFT(\<box>(empty \<longrightarrow> F))" 

definition halt_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "halt_d F \<equiv> LIFT(\<box>(empty = F))" 

definition initonly_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "initonly_d F \<equiv> LIFT(bi(empty = F))" 

definition keep_d :: "('a::world) formula \<Rightarrow> 'a formula "
where "keep_d F \<equiv> LIFT(ba(skip \<longrightarrow> F))" 

definition yields_d :: "('a::world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "yields_d F1 F2 \<equiv> LIFT(\<not>(F1;(\<not>F2)))" 

definition ifthenelse_d :: "('a::world) formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "ifthenelse_d F G H \<equiv> LIFT((F \<and> G) \<or> (\<not>F \<and> H) ) "

primrec power_d :: "('a::world) formula \<Rightarrow> nat \<Rightarrow> 'a formula "
where  pow_0  : "(power_d F 0) = LIFT(empty)"
     | pow_Suc: "(power_d F (Suc n)) = LIFT((F);(power_d F n)) "


syntax
 "_bm_d"         :: "lift \<Rightarrow> lift"             ("(bm _)" [88] 87)
 "_init_d"       :: "lift \<Rightarrow> lift"             ("(init _)" [88] 87)
 "_fin_d"        :: "lift \<Rightarrow> lift"             ("(fin _)" [88] 87)
 "_halt_d"       :: "lift \<Rightarrow> lift"             ("(halt _)" [88] 87)
 "_initonly_d"   :: "lift \<Rightarrow> lift"             ("(initonly _)" [88] 87)
 "_keep_d"       :: "lift \<Rightarrow> lift"             ("(keep _)" [88] 87)
 "_yields_d"     :: "[lift,lift] \<Rightarrow> lift"      ("(_ yields _)" [88,88] 87)
 "_ifthenelse_d" :: "[lift,lift,lift] \<Rightarrow> lift" ("(if\<^sub>i _ then _ else _ )"  [88,88,88] 87)
 "_power_d"      :: "[lift,nat] \<Rightarrow> lift"       ("(power _ _)" [88,88] 87) 
 
syntax (ASCII)
 "_bm_d"         :: "lift \<Rightarrow> lift"             ("(bm _)" [88] 87)
 "_init_d"       :: "lift \<Rightarrow> lift"             ("(init _)" [88] 87)
 "_fin_d"        :: "lift \<Rightarrow> lift"             ("(fin _)" [88] 87)
 "_halt_d"       :: "lift \<Rightarrow> lift"             ("(halt _)" [88] 87)
 "_initonly_d"   :: "lift \<Rightarrow> lift"             ("(initonly _)" [88] 87)
 "_keep_d"       :: "lift \<Rightarrow> lift"             ("(keep _)" [88] 87)
 "_yields_d"     :: "[lift,lift] \<Rightarrow> lift"      ("(_ yields _)" [88,88] 87)
 "_ifthenelse_d" :: "[lift,lift,lift] \<Rightarrow> lift" ("(if\<^sub>i _ then _ else _ )"  [88,88,88] 87)
 "_power_d"      :: "[lift,nat] \<Rightarrow> lift"       ("(power _ _)" [88,88] 87)

translations
 "_bm_d"         \<rightleftharpoons> "CONST bm_d"
 "_init_d"       \<rightleftharpoons> "CONST init_d"
 "_fin_d"        \<rightleftharpoons> "CONST fin_d"
 "_halt_d"       \<rightleftharpoons> "CONST halt_d"
 "_initonly_d"   \<rightleftharpoons> "CONST initonly_d" 
 "_keep_d"       \<rightleftharpoons> "CONST keep_d"
 "_yields_d"     \<rightleftharpoons> "CONST yields_d"
 "_ifthenelse_d" \<rightleftharpoons> "CONST ifthenelse_d" 
 "_power_d"      \<rightleftharpoons> "CONST power_d"

definition len_d :: "nat \<Rightarrow> ('a::world) formula "
where "len_d n \<equiv> LIFT(power skip n)"

definition powerstar_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "powerstar_d F \<equiv> LIFT(\<exists> k. power F k)"

syntax
 "_len_d"          :: "nat \<Rightarrow> lift"         ("(len _)" [88] 87) 
 "_powerstar_d"    :: "lift \<Rightarrow> lift"        ("(powerstar _)" [85] 85)

syntax (ASCII)
 "_len_d"          :: "nat \<Rightarrow> lift"         ("(len _)" [88] 87) 
 "_powerstar_d"    :: "lift \<Rightarrow> lift"        ("(powerstar _)" [85] 85)

translations
 "_len_d"          \<rightleftharpoons> "CONST len_d" 
 "_powerstar_d"    \<rightleftharpoons> "CONST powerstar_d"

definition chopstar_d :: "('a::world) formula \<Rightarrow> 'a formula"
where "chopstar_d F \<equiv> LIFT(powerstar (F \<and> more))"

syntax
 "_chopstar_d"     :: "lift \<Rightarrow> lift"        ("(_\<^sup>\<star>)" [85] 85)

syntax (ASCII)
 "_chopstar_d"     :: "lift \<Rightarrow> lift"        ("(chopstar _)" [85] 85)

translations
 "_chopstar_d"     \<rightleftharpoons> "CONST chopstar_d"


definition ifthen_d :: "('a::world) formula  \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "ifthen_d F G \<equiv> LIFT(if\<^sub>i F then G else #True ) "

definition while_d :: "('a::world) formula  \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "while_d F G \<equiv> LIFT( ( F \<and> G)\<^sup>\<star> \<and> (fin ((\<not>F))) ) "

syntax
 "_ifthen_d" :: "[lift,lift] \<Rightarrow> lift" ("(if\<^sub>i _ then _  )"  [88,88] 87)
 "_while_d"  :: "[lift,lift] \<Rightarrow> lift" ("(while _ do _  )"  [88,88] 87)

syntax (ASCII)
 "_ifthen_d" :: "[lift,lift] \<Rightarrow> lift" ("(if\<^sub>i _ then _  )"  [88,88] 87)
 "_while_d"  :: "[lift,lift] \<Rightarrow> lift" ("(while _ do _  )"  [88,88] 87)

translations
 "_ifthen_d" \<rightleftharpoons> "CONST ifthen_d"
 "_while_d"  \<rightleftharpoons> "CONST while_d"

definition repeat_d :: "('a::world) formula  \<Rightarrow> 'a formula \<Rightarrow> 'a formula "
where "repeat_d F G \<equiv> LIFT(F;while (\<not> G) do F ) "

syntax
 "_repeat_d"  :: "[lift,lift] \<Rightarrow> lift" ("(repeat _ until _  )"  [88,88] 87)

syntax (ASCII)
 "_repeat_d"  :: "[lift,lift] \<Rightarrow> lift" ("(repeat _ until _  )"  [88,88] 87)

translations
 "_repeat_d" \<rightleftharpoons> "CONST repeat_d"


definition next_assign_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "next_assign_d v e \<equiv> LIFT( v$ = e)" 

definition prev_assign_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "prev_assign_d v e \<equiv> LIFT( v! = e)"  

definition always_eq_d  :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "always_eq_d v e \<equiv> \<lambda> s. s \<Turnstile> \<box>($v = e)"  

definition temporal_assign_d  :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "temporal_assign_d v e \<equiv> \<lambda> s. s \<Turnstile> !v = e" 

definition gets_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "gets_d v e \<equiv> \<lambda> s. s \<Turnstile> keep( temporal_assign_d v e)" 

definition stable_d :: "('a::world,'b) stfun \<Rightarrow> 'a formula "
where " stable_d v \<equiv> \<lambda> s. s \<Turnstile> gets_d v (current_val_d v) "

definition padded_d :: "('a::world,'b) stfun \<Rightarrow> 'a formula "
where " padded_d v \<equiv> \<lambda> s. s \<Turnstile> (stable_d v);skip \<or> empty" 

definition padded_temp_assign_d :: "('a::world,'b) stfun \<Rightarrow> ('a,'b) formfun \<Rightarrow> 'a formula "
where "padded_temp_assign_d v e \<equiv> \<lambda> s. s \<Turnstile> (temporal_assign_d v e) \<and> (padded_d v)" 


syntax
 "_next_assign_d"        :: "[lift,lift] \<Rightarrow> lift" ("(_ := _)" [50,51] 50)
 "_prev_assign_d"        :: "[lift,lift] \<Rightarrow> lift" ("(_ =: _)" [50,51] 50)
 "_always_eq_d"          :: "[lift,lift] \<Rightarrow> lift" ("(_ \<approx> _)" [50,51] 50)
 "_temporal_assign_d"    :: "[lift,lift] \<Rightarrow> lift" ("(_ \<leftarrow> _)" [50,51] 50)
 "_gets_d"               :: "[lift,lift] \<Rightarrow> lift" ("(_ gets _)" [50,51] 50)
 "_stable_d"             :: "lift \<Rightarrow> lift"        ("(stable _)" [51] 50)
 "_padded_d"             :: "lift \<Rightarrow> lift"        ("(padded _)" [51] 50)
 "_padded_temp_assign_d" :: "[lift,lift] \<Rightarrow> lift" ("(_ <\<sim> _)" [50,51] 50)

syntax (ASCII)
 "_next_assign_d"        :: "[lift,lift] \<Rightarrow> lift" ("(_ := _)" [50,51] 50)
 "_prev_assign_d"        :: "[lift,lift] \<Rightarrow> lift" ("(_ =: _)" [50,51] 50)
 "_always_eq_d"          :: "[lift,lift] \<Rightarrow> lift" ("(_ alweqv _)" [50,51] 50)
 "_temporal_assign_d"    :: "[lift,lift] \<Rightarrow> lift" ("(_ <-- _)" [50,51] 50)
 "_gets_d"               :: "[lift,lift] \<Rightarrow> lift" ("(_ gets _)" [50,51] 50)
 "_stable_d"             :: "lift \<Rightarrow> lift"        ("(stable _)" [51] 50)
 "_padded_d"             :: "lift \<Rightarrow> lift"        ("(padded _)" [51] 50)
 "_padded_temp_assign_d" :: "[lift,lift] \<Rightarrow> lift" ("(_ <~ _)" [50,51] 50)

translations
 "_next_assign_d"        \<rightleftharpoons> "CONST next_assign_d"
 "_prev_assign_d"        \<rightleftharpoons> "CONST prev_assign_d"
 "_always_eq_d"          \<rightleftharpoons> "CONST always_eq_d"
 "_temporal_assign_d"    \<rightleftharpoons> "CONST temporal_assign_d"
 "_gets_d"               \<rightleftharpoons> "CONST gets_d"
 "_stable_d"             \<rightleftharpoons> "CONST stable_d"
 "_padded_d"             \<rightleftharpoons> "CONST padded_d"
 "_padded_temp_assign_d" \<rightleftharpoons> "CONST padded_temp_assign_d"


subsection \<open>Properties of Operators\<close>

text \<open> The following lemmas show that above operators have the expected semantics. \<close>

lemma skip_defs :
 "(w \<Turnstile> skip) = ( intlen w =1) "
by (simp add: skip_d_def)

lemma chop_defs :
 "(w \<Turnstile> F1 ; F2) = (\<exists> n . 0\<le>n \<and> n\<le>intlen w \<and> ((prefix n w)\<Turnstile> F1) \<and> ( (suffix n w) \<Turnstile> F2)  ) "
by (simp add: chop_d_def)

lemma sometimes_defs : 
  "(w \<Turnstile> \<diamond> F) = (\<exists> n. 0\<le>n \<and> n\<le>intlen w \<and> ((suffix n w) \<Turnstile> F))"
by (simp add: Semantics.sometimes_d_def chop_defs)

lemma always_defs : 
 " (w \<Turnstile> \<box> F) =(\<forall> n. 0\<le>n \<and> n\<le>intlen w \<longrightarrow> ((suffix n w) \<Turnstile> F))"
by (simp add: always_d_def sometimes_defs)

lemma di_defs : 
  "(w \<Turnstile> di F) = (\<exists> n. 0\<le>n \<and> n\<le>intlen w \<and> ((prefix n w) \<Turnstile> F))"
by (simp add: Semantics.di_d_def chop_defs)

lemma bi_defs : 
  "(w \<Turnstile> bi F) = (\<forall> n. 0\<le>n \<and> n\<le>intlen w \<longrightarrow> ((prefix n w) \<Turnstile> F))"
by (simp add: Semantics.bi_d_def di_defs)

lemma da_defs : 
  "(w \<Turnstile> da F) = (\<exists> n na. 0\<le>n \<and>  na+n \<le> intlen w \<and> ((sub n (na+n) w) \<Turnstile> F))"
using interval_prefix_length_good interval_suffix_length_good
by (simp add: Semantics.da_d_def chop_defs,
    smt add.commute add_diff_cancel_left' add_leD2 interval_sub_prefix_suffix_0 le_iff_add 
        nat_add_left_cancel_le zero_le)

lemma ba_defs :
 "(w \<Turnstile> ba F) = (\<forall> n na. 0\<le>n \<and>  na+n \<le> intlen w \<longrightarrow> ((sub n (na+n) w) \<Turnstile> F))"
by (simp add: ba_d_def da_defs)

lemma next_defs :
 " (w \<Turnstile> \<circle> F) = (intlen w > 0 \<and> ((suffix 1 w)\<Turnstile> F) ) "
using Suc_le_eq by (simp add: next_d_def chop_defs skip_defs, force)

lemma wnext_defs :
 "(w \<Turnstile> wnext F) = (intlen w = 0 \<or> ((suffix 1 w)\<Turnstile> F) ) "
by (simp add: wnext_d_def next_defs)

lemma prev_defs :
 " (w \<Turnstile> prev F) = (intlen w > 0 \<and> ((prefix ((intlen w)-1) w)\<Turnstile> F) ) "
by (simp add: prev_d_def chop_defs skip_defs)
 (metis One_nat_def Suc_leI diff_diff_cancel diff_is_0_eq' diff_le_self 
          interval_suffix_length_good neq0_conv zero_neq_one)

lemma wprev_defs :
 " (w \<Turnstile> wprev F) = (intlen w = 0 \<or> ((prefix ((intlen w)-1) w)\<Turnstile> F) ) "
by (metis (mono_tags, lifting) less_le prev_defs unl_lift wprev_d_def zero_le)

lemma more_defs :
 " (w \<Turnstile> more) = (intlen w>0)"
by (simp add: more_d_def next_defs)

lemma empty_defs :
 " (w \<Turnstile> empty) = (intlen w = 0)"
by (simp add: empty_d_def more_defs)

lemma init_defs :
 " (w \<Turnstile> init F) = ( (Interval.prefix 0 w) \<Turnstile> F )"
by (simp add: init_d_def empty_defs chop_defs) auto

lemma initalt_defs :
 " (w \<Turnstile> bi( empty \<longrightarrow> F)) = ( (Interval.prefix 0 w) \<Turnstile> F )" 
by (simp add: bi_defs empty_defs)

lemma fin_defs :
 " (w \<Turnstile> fin F) = ( (Interval.suffix (intlen w) w) \<Turnstile> F)"
by (simp add: fin_d_def empty_defs always_defs) 

lemma finalt_defs :
 " (w \<Turnstile> #True;(F \<and> empty)) = ( (Interval.suffix (intlen w) w) \<Turnstile> F)"
by (simp add: chop_defs empty_defs) fastforce

lemma halt_defs :
 " (w \<Turnstile> halt(F)) = (\<forall>n\<le>intlen w. (intlen w = n) = F (suffix n w))"
by (simp add: halt_d_def empty_defs always_defs)

lemma initonly_defs :
 " (w \<Turnstile> initonly(F)) = (\<forall>n\<le>intlen w. (n = 0) = F (prefix n w))"
by (simp add: initonly_d_def bi_defs empty_defs)

lemma ifthenelse_defs:
 " (w \<Turnstile> if\<^sub>i F then G else H) = 
   ( ((w \<Turnstile> F) \<and> (w \<Turnstile> G)) \<or> (( \<not>(w \<Turnstile> F) \<and> (w \<Turnstile> H))) )"
by (simp add: ifthenelse_d_def)

lemma len_defs :
 " (w \<Turnstile> len n) = (intlen w =n) "
by (induct n arbitrary: w, simp add: len_d_def empty_defs, 
    simp add: len_d_def chop_defs skip_defs) fastforce

lemma currentval_defs :
 " (s \<Turnstile> $v) = (v (nth s 0))"
by (simp add: current_val_d_def)

lemma nextval_defs :
 "(s \<Turnstile> v$) = (if intlen s >0 then (v (nth s 1)) else (\<some> x. x=x)) "
by (simp add: next_val_d_def)

lemma finval_defs :
 " (s \<Turnstile> !v) = (v (nth s (intlen s)))"
by (simp add: fin_val_d_def)

lemma penultval_defs :
 " (s \<Turnstile> v!) = (if intlen s >0 then (v (nth s ((intlen s)-1))) else (\<some> x. x=x))"
by (simp add: penult_val_d_def) 

lemma next_assign_defs :
 " intlen s >0 \<Longrightarrow> (s \<Turnstile> v := e) = v (Interval.nth s 1) = e s "
by (auto simp: next_assign_d_def next_val_d_def)

lemma prev_assign_defs :
 " intlen s >0 \<Longrightarrow> (s \<Turnstile> v =: e) = v (Interval.nth s ((intlen s)-1)) = e s "
by (auto simp: prev_assign_d_def penult_val_d_def)

lemma always_eqv_defs :
 " (s \<Turnstile> v \<approx> e) = (\<forall> i\<le> intlen s. v (Interval.nth s i) = e (suffix i s))"
by (simp add: always_eq_d_def always_defs current_val_d_def)

lemma temporal_assign_defs :
 " (s \<Turnstile> v \<leftarrow> e) = (v (Interval.nth s (intlen s)) = e s)"
by (simp add: temporal_assign_d_def fin_val_d_def)

lemma gets_defs :
 " (s \<Turnstile> v gets e) = (\<forall> i < intlen s. v (Interval.nth s (Suc i)) = e (sub i (i+1) s) )"
using Suc_le_eq 
by (simp add: gets_d_def keep_d_def ba_defs skip_defs sub_def temporal_assign_defs)

lemma stable_defs_help:
 "(\<forall>i<intlen s. v (Interval.nth s (Suc i)) = v (Interval.nth s i)) = 
  (\<forall>i\<le>intlen s. v (Interval.nth s i) = v (Interval.nth s 0))"
proof 
  (induct s)
  case (St x)
  then show ?case by simp
  next
  case (Cons x1a s)
  then show ?case 
  by (smt Suc_lessI interval_nth_Suc intlen.simps(2) le_SucE le_neq_implies_less le_simps(1) 
          less_Suc_eq plus_1_eq_Suc zero_less_Suc)
qed

lemma stable_defs:
 "(s \<Turnstile> stable v) = (\<forall> i\<le>intlen s. (v (nth s i)) = (v (nth s 0)))" 
by (simp add: stable_d_def gets_defs current_val_d_def sub_def stable_defs_help)

lemma padded_defs :
 " (s \<Turnstile> padded v) = ((\<forall> i< intlen s. (v (nth s i)) = (v (nth s 0))) \<or> intlen s =0) "
by (simp add: padded_d_def stable_defs chop_d_def skip_defs empty_defs interval_suffix_length,
    smt Suc_leI Suc_pred diff_diff_cancel interval_intlen_gr_zero le_neq_implies_less le_simps(1) 
        less_Suc_eq)

lemma padded_temporal_assign_defs :
 " (s \<Turnstile> v <\<sim> e) = 
  ((s \<Turnstile> padded v) \<and>
    (v (Interval.nth s (intlen s)) = e s ))"
by (simp add: padded_temp_assign_d_def padded_defs temporal_assign_defs, auto)

lemma linalw:
  "a \<le> b \<and> b\<le>intlen w \<and> ((suffix a w  )\<Turnstile> \<box> A) \<longrightarrow> ((suffix b w ) \<Turnstile> \<box> A)"
by (simp add: always_defs,
    smt add.assoc add.commute interval_suffix_length_good le_add_diff_inverse le_trans 
        ordered_cancel_comm_monoid_diff_class.le_diff_conv2)
 

subsection \<open>Soundness of Finite ITL Axioms\<close>

subsubsection \<open>ChopAssoc\<close>

lemma ChopAssocSemHelp:
 "(\<exists>i ia . i\<le>intlen \<sigma> \<and> ia \<le> intlen \<sigma> -i \<and> (prefix i \<sigma> \<Turnstile> f) \<and>
    (prefix ia (suffix i \<sigma>) \<Turnstile> g) \<and> (suffix (ia + i) \<sigma> \<Turnstile> h)) =
  (\<exists>j ja . j \<le>intlen \<sigma> \<and> ja \<le> j \<and> (prefix ja (prefix j \<sigma>) \<Turnstile> f) \<and> 
    (suffix ja (prefix j \<sigma>) \<Turnstile> g) \<and> (suffix j \<sigma> \<Turnstile> h))" 
by (smt Nat.le_diff_conv2 add_diff_cancel_left' interval_pref_pref_3 interval_suffix_prefix_swap 
        le_add1 le_add_diff_inverse2 le_trans)

lemma ChopAssocSemHelp2:
 " (\<sigma> \<Turnstile>  f ; (g ; h)) = (\<sigma> \<Turnstile> (f;g);h)"
proof -
 have " (\<sigma> \<Turnstile>  f ; (g ; h)) =
        ((\<exists>i\<le>intlen \<sigma>. (prefix i \<sigma> \<Turnstile> f) \<and> (\<exists>ia\<le>intlen (suffix i \<sigma>). 
          (prefix ia (suffix i \<sigma>) \<Turnstile> g) \<and> (suffix (ia + i) \<sigma> \<Turnstile> h))))" 
 by (simp add: chop_defs)
 also have " ... = 
             (\<exists>i ia . i\<le>intlen \<sigma> \<and> ia \<le> intlen \<sigma> -i \<and> (prefix i \<sigma> \<Turnstile> f) \<and>
         (prefix ia (suffix i \<sigma>) \<Turnstile> g) \<and> (suffix (ia + i) \<sigma> \<Turnstile> h))" 
 by fastforce
 also have " ... =
            (\<exists>j ja . j \<le>intlen \<sigma> \<and> ja \<le> j \<and> (prefix ja (prefix j \<sigma>) \<Turnstile> f) \<and>
            (suffix ja (prefix j \<sigma>) \<Turnstile> g) \<and> (suffix j \<sigma> \<Turnstile> h))" 
 using ChopAssocSemHelp[of \<sigma> f g h]  by blast
 also have " ... =
            (\<exists>i\<le>intlen \<sigma>. (\<exists>ia\<le>intlen (prefix i \<sigma>).  (prefix ia (prefix i \<sigma>) \<Turnstile> f) \<and>
             (suffix ia (prefix i \<sigma>) \<Turnstile> g)) \<and> (suffix i \<sigma> \<Turnstile> h)) " 
 by fastforce
 also have " ... =
             (\<sigma> \<Turnstile> (f;g);h)" by (simp add: chop_defs)
 finally show "(\<sigma> \<Turnstile>  f ; (g ; h)) = (\<sigma> \<Turnstile> (f;g);h) " .
qed

lemma ChopAssocSem:
  " (\<sigma> \<Turnstile>  f ; (g ; h) =  (f;g);h)"
using ChopAssocSemHelp2 using unl_lift2 by blast

subsubsection \<open>OrChopImp\<close>

lemma OrChopImpSem:
  " (\<sigma> \<Turnstile>  ( f \<or> g);h   \<longrightarrow>  f;h \<or> g;h ) "
by (simp add: chop_defs) blast

subsubsection \<open>ChopOrImp\<close>

lemma ChopOrImpSem:
  " (\<sigma> \<Turnstile> f;(g \<or> h) \<longrightarrow>   f;g \<or>  f;h ) "
by (simp add: chop_defs) blast

subsubsection \<open>EmptyChop\<close>

lemma EmptyChopSem:
  " (\<sigma> \<Turnstile> empty ; f = f ) "
by (simp add: empty_defs chop_defs) auto

subsubsection \<open>ChopEmpty\<close>

lemma ChopEmptySem:
  " (\<sigma> \<Turnstile> f;empty = f ) "
by (simp add: empty_defs chop_defs) auto

subsubsection \<open>StateImpBi\<close>

lemma StateImpBiSem:
  " (\<sigma> \<Turnstile> init f \<longrightarrow>   bi (init f) ) "
by (simp add: init_defs bi_defs)

subsubsection \<open>NextImpNotNextNot\<close>

lemma NextImpNotNextNotSem:
  " (\<sigma> \<Turnstile> \<circle> f \<longrightarrow>  \<not> (\<circle> (\<not> f)) ) "
by (simp add: next_defs) 

subsubsection \<open>BiBoxChopImpChop\<close>

lemma BiBoxChopImpChopSem:
  " (\<sigma> \<Turnstile> bi ( f \<longrightarrow> f1) \<and> \<box>(g \<longrightarrow> g1) \<longrightarrow> f;g \<longrightarrow> f1;g1 ) "
by (simp add: bi_defs always_defs chop_defs) fastforce

subsubsection \<open>BoxInduct\<close>

lemma box_induct_help_1 :
  "(\<sigma> \<Turnstile> f) \<and> (\<forall>i. Suc 0 \<le> intlen \<sigma> - i \<longrightarrow>
         i \<le> intlen \<sigma> \<longrightarrow> (suffix i \<sigma> \<Turnstile> f) \<longrightarrow> (suffix (Suc i) \<sigma> \<Turnstile> f)) 
     \<Longrightarrow> (\<forall> j. j \<le> intlen \<sigma> \<longrightarrow> (suffix j \<sigma> \<Turnstile> f)) "
proof
    fix j
    show " (\<sigma> \<Turnstile> f) \<and> (\<forall>i. Suc 0 \<le> intlen \<sigma> - i \<longrightarrow>
         i \<le> intlen \<sigma> \<longrightarrow> (suffix i \<sigma> \<Turnstile> f) \<longrightarrow> (suffix (Suc i) \<sigma> \<Turnstile> f)) 
          \<Longrightarrow>  j \<le> intlen \<sigma> \<longrightarrow> (suffix j \<sigma> \<Turnstile> f)" 
     proof
       (induct j arbitrary: \<sigma>)
       case 0
       then show ?case by simp
       next
       case (Suc j)
       then show ?case 
       by (metis Nat.le_diff_conv2 One_nat_def Suc_eq_plus1_left Suc_leD)
       qed
qed

lemma BoxInductSem:
  " (\<sigma> \<Turnstile> \<box> (f \<longrightarrow> wnext f) \<and> f \<longrightarrow> \<box> f)"
proof -
 have 1: "(\<sigma> \<Turnstile> \<box> (f \<longrightarrow> wnext f) \<and> f \<longrightarrow> \<box> f) =
          ((\<forall>n\<le>intlen \<sigma>. f (suffix n \<sigma>) \<longrightarrow> intlen \<sigma> = n \<or> f (suffix (Suc n) \<sigma>)) \<and> f \<sigma> \<longrightarrow> 
           (\<forall>n\<le>intlen \<sigma>. f (suffix n \<sigma>)))"
       by (simp add: always_defs wnext_defs)
 from 1 show ?thesis using box_induct_help_1 
  by (metis One_nat_def diff_self_eq_0 not_one_le_zero)
qed
 
subsubsection \<open>ChopStarEqv\<close>

lemma ChopExist:
 "\<turnstile>  (\<exists> k. f;g k) = f;(\<exists> k. g k)"
by (simp add: chop_defs Valid_def, auto)

lemma ExistChop:
 "\<turnstile> (\<exists> k. (g k);f) =  (\<exists> k. g k);f"
by (simp add: chop_defs Valid_def, auto)

lemma powersem1:
 "(\<sigma> \<Turnstile>  (\<exists> k. power f k) = ( empty \<or> (\<exists> k. power f (Suc k)))) "
by (smt not0_implies_Suc pow_0 unl_Rex unl_lift2)

lemma powersem:
 " \<turnstile> (\<exists> k. power f k) = (  empty \<or> (f);(\<exists> k. (power f k)))"
proof -
 have 1: "\<turnstile> (\<exists> k. power f k) = ( empty \<or> (\<exists> k. power f (Suc k))) "
 using powersem1 by blast
 have 2: "\<turnstile> (\<exists> k. power f (Suc k)) = (\<exists> k. (f);power f k) "
 by simp
 have 3: "\<turnstile> (\<exists> k. (f);(power f k)) = (f);(\<exists> k. (power f k)) " 
 using ChopExist by blast
 from 1 2 3 show ?thesis by fastforce
qed

lemma PowerstarEqvSem:
  " (\<sigma> \<Turnstile> (powerstar f) =  (empty \<or>  f;(powerstar f) ))"
proof -
 have 1: "(\<sigma> \<Turnstile> (powerstar f)) = 
          ( \<sigma> \<Turnstile> (\<exists> k. power f k)) "
 by (simp add: powerstar_d_def)
 have 2: "( \<sigma> \<Turnstile> (\<exists> k. power f k)) =
          (\<sigma> \<Turnstile> (empty \<or> f;(\<exists> k. (power f k))))"
 using powersem by (metis inteq_reflection)
 from 1 2 show ?thesis by (simp add: powerstar_d_def)
qed
 
lemma ChopstarEqvSem:
  " (\<sigma> \<Turnstile> f\<^sup>\<star> =  (empty \<or>  (f \<and> more); f\<^sup>\<star>) )"
by (metis PowerstarEqvSem chopstar_d_def)


subsection \<open>Quantification over State (Flexible) Variables\<close>

text \<open>
  The hidden state approach, as used in the embedding of TLA in 
  Isabelle/HOL TLA embedding \cite{Merz98,GrovMerz}, is used. 
  Here \cite{Merz98,GrovMerz}, a state space is defined by its projections, 
  and everything else is
  unknown. Thus, a variable is a projection of the state space, and has the same
  type as a state function. Moreover, strong typing is achieved, since the projection
  function may have any result type. To achieve this, the state space is represented
  by an undefined type, which is an instance of the \<open>world\<close> class to enable
  use with the \<open>Intensional\<close> theory.
\<close>

typedecl state

instance state :: world ..

type_synonym 'a statefun = "(state,'a) stfun"
type_synonym statepred   = "bool statefun"
type_synonym 'a tempfun  = "(state,'a) formfun"
type_synonym temporal    = "state formula"

text \<open> 
  Similar to \cite{Merz98,GrovMerz} we define a
  state to be an anonymous type whose only purpose is to provide Skolem constants.
  Similarly, we do not define a type of state variables separate from that
  of arbitrary state functions, again in order to simplify the definition
  of flexible quantification later on. Nevertheless, we need to distinguish
  state variables. Note we deviate from \cite{Merz98,GrovMerz} in that we do not use 
  axioms but use definitions and lemmas.
\<close>


subsection \<open>Temporal Quantifiers\<close>

definition exist_state_d :: " ('a statefun \<Rightarrow>  temporal )\<Rightarrow> temporal" (binder "Eex " 10)
where "exist_state_d  F  \<equiv> (\<lambda>s.  (\<exists> x.  s \<Turnstile> F x ))"

syntax
 "_Eex" :: "[idts, lift] \<Rightarrow> lift"      ("(3\<exists>\<exists> _./ _)" [0,10] 10)

translations
 "_Eex v A"  == "Eex v. A" 

definition forall_state_d :: " ('a statefun \<Rightarrow>  temporal )\<Rightarrow> temporal" (binder "Aall " 10)
where "forall_state_d F \<equiv> LIFT(\<not>(\<exists>\<exists> x. \<not>(F x)))"

syntax 
 "_Aall" :: "[idts, lift] \<Rightarrow> lift"      ("(3\<forall>\<forall> _./ _)" [0,10] 10)

translations
 "_Aall v A" == "Aall v. A"



end
